/*
 * Copyright (c) 2017 SpaceToad and the BuildCraft team
 * This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0. If a copy of the MPL was not
 * distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/
 */

package buildcraft.lib.misc;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.stream.Collector;
import java.util.stream.Collectors;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import net.minecraft.block.Block;
import net.minecraft.block.state.IBlockState;
import net.minecraft.item.Item;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.util.NonNullList;

import net.minecraftforge.oredict.OreDictionary;

import buildcraft.api.items.IList;
import buildcraft.api.recipes.IngredientStack;
import buildcraft.api.recipes.StackDefinition;

/** Provides various utils for interacting with {@link ItemStack}, and multiples. */
public class StackUtil {

    /** A non-null version of {@link ItemStack#EMPTY}. When the original field adds an @Nonnull annotation this should
     * be inlined. */
    // Actually the entire MC
    @Nonnull
    public static final ItemStack EMPTY;

    /**
     * Registry of additional rules for {@link #isMatchingItem}.
     */
    private static final Map<Item, List<StackMatchingPredicate>> matchingPredicates = new HashMap<>();

    static {
        ItemStack stack = ItemStack.field_190927_a;
        if (stack == null) throw new NullPointerException("Empty ItemStack was null!");
        EMPTY = stack;
    }

    /** Checks to see if the two input stacks are equal in all but stack size. Note that this doesn't check anything
     * todo with stack size, so if you pass in two stacks of 64 cobblestone this will return true. If you pass in null
     * (at all) then this will only return true if both are null. */
    public static boolean canMerge(@Nonnull ItemStack a, @Nonnull ItemStack b) {
        // Checks item, damage
        if (!ItemStack.func_179545_c(a, b)) {
            return false;
        }
        // checks tags and caps
        return ItemStack.func_77970_a(a, b);
    }

    /** Attempts to get an item stack that might place down the given blockstate. Obviously this isn't perfect, and so
     * cannot be relied on for anything more than simple blocks. */
    @Nonnull
    public static ItemStack getItemStackForState(IBlockState state) {
        Block b = state.func_177230_c();
        ItemStack stack = new ItemStack(b);
        if (stack.func_190926_b()) {
            return StackUtil.EMPTY;
        }
        if (stack.func_77981_g()) {
            stack = new ItemStack(stack.func_77973_b(), 1, b.func_176201_c(state));
        }
        return stack;
    }

    /** Checks to see if the given required stack is contained fully in the given container stack. */
    public static boolean contains(@Nonnull ItemStack required, @Nonnull ItemStack container) {
        if (canMerge(required, container)) {
            return container.func_190916_E() >= required.func_190916_E();
        }
        return false;
    }

    /** Checks to see if the given required stack is contained fully in a single stack in a list. */
    public static boolean contains(@Nonnull ItemStack required, Collection<ItemStack> containers) {
        for (ItemStack possible : containers) {
            if (possible == null) {
                // Use an explicit null check here as the collection doesn't have @Nonnull applied to its type
                throw new NullPointerException("Found a null itemstack in " + containers);
            }
            if (contains(required, possible)) {
                return true;
            }
        }
        return false;
    }

    /** Checks that passed stack meets stack definition requirements */
    public static boolean contains(@Nonnull StackDefinition stackDefinition, @Nonnull ItemStack stack) {
        return !stack.func_190926_b() && stackDefinition.filter.matches(stack) && stack.func_190916_E() >= stackDefinition.count;
    }

    /** Checks that passed stack definition acceptable for stack collection */
    public static boolean contains(@Nonnull StackDefinition stackDefinition, @Nonnull NonNullList<ItemStack> stacks) {
        return stacks.stream().anyMatch((stack) -> contains(stackDefinition, stack));
    }

    /** Checks that passed stack meets stack definition requirements */
    public static boolean contains(@Nonnull IngredientStack ingredientStack, @Nonnull ItemStack stack) {
        return !stack.func_190926_b() && ingredientStack.ingredient.apply(stack) && stack.func_190916_E() >= ingredientStack.count;
    }

    /** Checks that passed stack definition acceptable for stack collection */
    public static boolean contains(@Nonnull IngredientStack ingredientStack, @Nonnull NonNullList<ItemStack> stacks) {
        return stacks.stream().anyMatch((stack) -> contains(ingredientStack, stack));
    }

    /** Checks to see if the given required stacks are all contained within the collection of containers. Note that this
     * assumes that all of the required stacks are different. */
    public static boolean containsAll(Collection<ItemStack> required, Collection<ItemStack> containers) {
        for (ItemStack req : required) {
            if (req == null) {
                // Use an explicit null check here as the collection doesn't have @Nonnull applied to its type
                throw new NullPointerException("Found a null itemstack in " + containers);
            }
            if (req.func_190926_b()) continue;
            if (!contains(req, containers)) {
                return false;
            }
        }
        return true;
    }

    public static NBTTagCompound stripNonFunctionNbt(@Nonnull ItemStack from) {
        NBTTagCompound nbt = NBTUtilBC.getItemData(from).func_74737_b();
        if (nbt.func_186856_d() == 0) {
            return nbt;
        }
        nbt.func_82580_o("_data");
        // TODO: Remove all of the non functional stuff (name, desc, etc)
        return nbt;
    }

    public static boolean doesStackNbtMatch(@Nonnull ItemStack target, @Nonnull ItemStack with) {
        NBTTagCompound nbtTarget = stripNonFunctionNbt(target);
        NBTTagCompound nbtWith = stripNonFunctionNbt(with);
        return nbtTarget.equals(nbtWith);
    }

    public static boolean doesEitherStackMatch(@Nonnull ItemStack stackA, @Nonnull ItemStack stackB) {
        return OreDictionary.itemMatches(stackA, stackB, false) || OreDictionary.itemMatches(stackB, stackA, false);
    }

    public static boolean canStacksOrListsMerge(@Nonnull ItemStack stack1, @Nonnull ItemStack stack2) {
        if (stack1.func_190926_b() || stack2.func_190926_b()) {
            return false;
        }

        if (stack1.func_77973_b() instanceof IList) {
            IList list = (IList) stack1.func_77973_b();
            return list.matches(stack1, stack2);
        } else if (stack2.func_77973_b() instanceof IList) {
            IList list = (IList) stack2.func_77973_b();
            return list.matches(stack2, stack1);
        }

        return stack1.func_77969_a(stack2) && ItemStack.func_77970_a(stack1, stack2);
    }

    /** This doesn't take into account stack sizes.
     * 
     * @param filterOrList The exact itemstack to test, or an item that implements {@link IList} to test against.
     * @param test The stack to test for equality
     * @return True if they matched according to the above definitions, or false if theydidn't, or either was empty. */
    public static boolean matchesStackOrList(@Nonnull ItemStack filterOrList, @Nonnull ItemStack test) {
        if (filterOrList.func_190926_b() || test.func_190926_b()) {
            return false;
        }
        if (filterOrList.func_77973_b() instanceof IList) {
            IList list = (IList) filterOrList.func_77973_b();
            return list.matches(filterOrList, test);
        }
        return canMerge(filterOrList, test);
    }

    /** Merges mergeSource into mergeTarget
     *
     * @param mergeSource - The stack to merge into mergeTarget, this stack is not modified
     * @param mergeTarget - The target merge, this stack is modified if doMerge is set
     * @param doMerge - To actually do the merge
     * @return The number of items that was successfully merged. */
    public static int mergeStacks(@Nonnull ItemStack mergeSource, @Nonnull ItemStack mergeTarget, boolean doMerge) {
        if (!canMerge(mergeSource, mergeTarget)) {
            return 0;
        }
        int mergeCount = Math.min(mergeTarget.func_77976_d() - mergeTarget.func_190916_E(), mergeSource.func_190916_E());
        if (mergeCount < 1) {
            return 0;
        }
        if (doMerge) {
            mergeTarget.func_190920_e(mergeTarget.func_190916_E() + mergeCount);
        }
        return mergeCount;
    }

    /* ITEM COMPARISONS */
    /** Determines whether the given ItemStack should be considered equivalent for crafting purposes.
     *
     * @param base The stack to compare to.
     * @param comparison The stack to compare.
     * @param oreDictionary true to take the Forge OreDictionary into account.
     * @return true if comparison should be considered a crafting equivalent for base. */
    public static boolean isCraftingEquivalent(@Nonnull ItemStack base, @Nonnull ItemStack comparison,
        boolean oreDictionary) {
        if (isMatchingItem(base, comparison, true, false)) {
            return true;
        }

        if (oreDictionary) {
            int[] idBase = OreDictionary.getOreIDs(base);
            if (idBase.length > 0) {
                for (int id : idBase) {
                    for (ItemStack itemstack : OreDictionary.getOres(OreDictionary.getOreName(id))) {
                        if (comparison.func_77973_b() == itemstack.func_77973_b()
                            && (itemstack.func_77952_i() == OreDictionary.WILDCARD_VALUE
                                || comparison.func_77952_i() == itemstack.func_77952_i())) {
                            return true;
                        }
                    }
                }
            }
        }

        return false;
    }

    public static boolean isCraftingEquivalent(int[] oreIDs, ItemStack comparison) {
        if (oreIDs.length > 0) {
            for (int id : oreIDs) {
                for (ItemStack itemstack : OreDictionary.getOres(OreDictionary.getOreName(id))) {
                    if (comparison.func_77973_b() == itemstack.func_77973_b()
                        && (itemstack.func_77952_i() == OreDictionary.WILDCARD_VALUE
                            || comparison.func_77952_i() == itemstack.func_77952_i())) {
                        return true;
                    }
                }
            }
        }

        return false;
    }

    public static boolean isMatchingItemOrList(final ItemStack base, final ItemStack comparison) {
        if (base.func_190926_b() || comparison.func_190926_b()) {
            return false;
        }

        if (base.func_77973_b() instanceof IList) {
            IList list = (IList) base.func_77973_b();
            return list.matches(base, comparison);
        } else if (comparison.func_77973_b() instanceof IList) {
            IList list = (IList) comparison.func_77973_b();
            return list.matches(comparison, base);
        }

        return isMatchingItem(base, comparison, true, false);
    }

    /** Compares item id, damage and NBT. Accepts wildcard damage. Ignores damage entirely if the item doesn't have
     * subtypes.
     *
     * @param base The stack to compare to.
     * @param comparison The stack to compare.
     * @return true if id, damage and NBT match. */
    public static boolean isMatchingItem(final @Nonnull ItemStack base, final @Nonnull ItemStack comparison) {
        return isMatchingItem(base, comparison, true, true);
    }

    /** This variant also checks damage for damaged items. */
    public static boolean isEqualItem(final @Nonnull ItemStack base, final @Nonnull ItemStack comparison) {
        if (isMatchingItem(base, comparison, false, true)) {
            return isWildcard(base) || isWildcard(comparison) || base.func_77952_i() == comparison.func_77952_i();
        } else {
            return false;
        }
    }

    /** Compares item id, and optionally damage and NBT. Accepts wildcard damage. Ignores damage entirely if the item
     * doesn't have subtypes.
     *
     * @param base ItemStack
     * @param comparison ItemStack
     * @param matchDamage
     * @param matchNBT
     * @return true if matches */
    public static boolean isMatchingItem(@Nonnull final ItemStack base, @Nonnull final ItemStack comparison,
        final boolean matchDamage, final boolean matchNBT) {
        if (base.func_190926_b() || comparison.func_190926_b()) {
            return false;
        }

        if (base.func_77973_b() != comparison.func_77973_b()) {
            return false;
        }
        if (matchDamage && base.func_77981_g()) {
            if (!isWildcard(base) && !isWildcard(comparison)) {
                if (base.func_77952_i() != comparison.func_77952_i()) {
                    return false;
                }
            }
        }
        if (matchNBT) {
            NBTTagCompound baseTag = base.func_77978_p();
            if (baseTag != null && !baseTag.equals(comparison.func_77978_p())) {
                return false;
            }
        } else {
            List<StackMatchingPredicate> predicates = matchingPredicates.getOrDefault(base.func_77973_b(), Collections.emptyList());
            for (StackMatchingPredicate predicate : predicates) {
                if (!predicate.isMatching(base, comparison)) {
                    return false;
                }
            }
        }
        return true;
    }

    /**
     * Registers a predicate, that will be used in {@link #isMatchingItem} as an additional comparison rule.
     * If any of registered predicates will return false, then the function will also return false.
     * It can be helpful, if item stacks are clearly not the same, but have common {@link Item} instance.
     * @param forItem {@link Item} instance for which to register the rule.
     * @param predicate predicate to register.
     */
    public static void registerMatchingPredicate(@Nonnull Item forItem, @Nonnull StackMatchingPredicate predicate) {
        List<StackMatchingPredicate> predicates = matchingPredicates.computeIfAbsent(forItem, (item) -> new ArrayList<>());
        predicates.add(predicate);
    }

    /** Checks to see if the given {@link ItemStack} is considered to be a wildcard stack - that is any damage value on
     * the stack will be considered the same as this for recipe purposes.
     * 
     * @param stack The stack to check
     * @return True if the stack is a wildcard, false if not. */
    public static boolean isWildcard(@Nonnull ItemStack stack) {
        return isWildcard(stack.func_77952_i());
    }

    /** Checks to see if the given {@link ItemStack} is considered to be a wildcard stack - that is any damage value on
     * the stack will be considered the same as this for recipe purposes.
     * 
     * @param damage The damage to check
     * @return True if the damage does specify a wildcard, false if not. */
    public static boolean isWildcard(int damage) {
        return damage == -1 || damage == OreDictionary.WILDCARD_VALUE;
    }

    /** @return An empty, nonnull list that cannot be modified (as it cannot be expanded and it has a size of 0) */
    public static NonNullList<ItemStack> listOf() {
        return NonNullList.func_191197_a(0, EMPTY);
    }

    /** Creates a {@link NonNullList} of {@link ItemStack}'s with the elements given in the order that they are given.
     * 
     * @param stacks The stacks to put into a list
     * @return A {@link NonNullList} of all the given items. Note that the returned list of of a specified size, and
     *         cannot be expanded. */
    public static NonNullList<ItemStack> listOf(ItemStack... stacks) {
        switch (stacks.length) {
            case 0:
                return listOf();
            case 1:
                return NonNullList.func_191197_a(1, stacks[0]);
            default:
        }
        NonNullList<ItemStack> list = NonNullList.func_191197_a(stacks.length, EMPTY);
        for (int i = 0; i < stacks.length; i++) {
            list.set(i, stacks[i]);
        }
        return list;
    }

    /** Takes a {@link Nullable} {@link Object} and checks to make sure that it is really {@link Nonnull}, like it is
     * everywhere else in the codebase. This is only required if some classes do not use the {@link Nonnull} annotation
     * on return values.
     * 
     * @param obj The (potentially) null object.
     * @return A {@link Nonnull} object, which will be the input object
     * @throws NullPointerException if the input object was actually null (Although this should never happen, this is
     *             more to catch bugs in dev.) */
    @Nonnull
    public static <T> T asNonNull(@Nullable T obj) {
        if (obj == null) {
            throw new NullPointerException("Object was null!");
        }
        return obj;
    }

    @Nonnull
    public static <T> T asNonNullSoft(@Nullable T obj, @Nonnull T fallback) {
        if (obj == null) {
            return fallback;
        } else {
            return obj;
        }
    }

    @Nonnull
    public static ItemStack asNonNullSoft(@Nullable ItemStack stack) {
        return asNonNullSoft(stack, EMPTY);
    }

    /** @return A {@link Collector} that will collect the input elements into a {@link NonNullList} */
    public static <E> Collector<E, ?, NonNullList<E>> nonNullListCollector() {
        return Collectors.toCollection(NonNullList::func_191196_a);
    }

    /** Computes a hash code for the given {@link ItemStack}. This is based off of {@link ItemStack#serializeNBT()},
     * except if {@link ItemStack#isEmpty()} returns true, in which case the hash will be 0. */
    public static int hash(@Nonnull ItemStack stack) {
        if (stack.func_190926_b()) {
            return 0;
        }
        if (!stack.func_77942_o()) {
            return Objects.hash(stack.func_77973_b(), stack.func_77960_j());
        }
        return stack.serializeNBT().hashCode();
    }

    public static NonNullList<ItemStack> mergeSameItems(List<ItemStack> items) {
        NonNullList<ItemStack> stacks = NonNullList.func_191196_a();
        for (ItemStack toAdd : items) {
            boolean found = false;
            for (ItemStack stack : stacks) {
                if (canMerge(stack, toAdd)) {
                    stack.func_190917_f(toAdd.func_190916_E());
                    found = true;
                }
            }
            if (!found) {
                stacks.add(toAdd.func_77946_l());
            }
        }
        return stacks;
    }
}
