/*
 * Copyright (c) 2017 SpaceToad and the BuildCraft team
 * This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0. If a copy of the MPL was not
 * distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/
 */

package buildcraft.lib.misc.data;

import java.util.Iterator;

import javax.annotation.Nonnull;

import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.EnumFacing.AxisDirection;
import net.minecraft.util.math.BlockPos;

import buildcraft.api.core.IBox;

import buildcraft.lib.misc.NBTUtilBC;
import buildcraft.lib.misc.StringUtilBC;
import buildcraft.lib.misc.VecUtil;

public class BoxIterator implements Iterator<BlockPos> {
    @Nonnull
    private final BlockPos min, max;
    private final boolean invert, repeat;// TODO: remove repeat if its not used in the future
    private AxisOrder order;
    private BlockPos current;
    private boolean hasRepeated = false;

    public BoxIterator(IBox box, AxisOrder order, boolean invert) {
        this(box.min(), box.max(), order, invert);
    }

    public BoxIterator(BlockPos min, BlockPos max, AxisOrder order, boolean invert) {
        this(min, max, invert, false, order, null);
    }

    private BoxIterator(BlockPos min, BlockPos max, boolean invert, boolean repeat, AxisOrder order, BlockPos current) {
        if (min == null) throw new NullPointerException("min");
        if (max == null) throw new NullPointerException("max");
        if (order == null) throw new NullPointerException("order");
        this.min = min;
        this.max = max;
        this.invert = invert;
        this.repeat = repeat;
        this.order = order;
        this.current = current == null ? getStart() : current;
    }

    public static BoxIterator readFromNbt(NBTTagCompound nbt) {
        BlockPos min = NBTUtilBC.readBlockPos(nbt.func_74781_a("min"));
        BlockPos max = NBTUtilBC.readBlockPos(nbt.func_74781_a("max"));
        boolean invert = nbt.func_74767_n("invert");
        boolean repeat = false;
        AxisOrder order = AxisOrder.readNbt(nbt.func_74775_l("order"));
        BlockPos current = NBTUtilBC.readBlockPos(nbt.func_74781_a("current"));
        if (min == null || max == null || order == null) {
            return null;
        }
        return new BoxIterator(min, max, invert, repeat, order, current);
    }

    public NBTTagCompound writeToNbt() {
        NBTTagCompound nbt = new NBTTagCompound();
        nbt.func_74782_a("min", NBTUtilBC.writeBlockPos(min));
        nbt.func_74782_a("max", NBTUtilBC.writeBlockPos(max));
        nbt.func_74757_a("invert", invert);
        // repeat
        nbt.func_74782_a("order", order.writeNBT());
        if (current != null) {
            nbt.func_74782_a("current", NBTUtilBC.writeBlockPos(current));
        }
        return nbt;
    }

    private BlockPos getStart() {
        BlockPos pos = BlockPos.field_177992_a;
        pos = replace(pos, order.first);
        pos = replace(pos, order.second);
        return replace(pos, order.third);
    }

    private BlockPos replace(BlockPos toReplace, EnumFacing facing) {
        BlockPos with = facing.func_176743_c() == AxisDirection.POSITIVE ? min : max;
        return VecUtil.replaceValue(toReplace, facing.func_176740_k(), VecUtil.getValue(with, facing.func_176740_k()));
    }

    public BlockPos getCurrent() {
        return current;
    }

    @Nonnull
    public BlockPos getMin() {
        return min;
    }

    @Nonnull
    public BlockPos getMax() {
        return max;
    }

    public boolean isInvert() {
        return invert;
    }

    public boolean isRepeat() {
        return repeat;
    }

    public AxisOrder getOrder() {
        return order;
    }

    @Override
    public String toString() {
        return "{BoxIterator [" + StringUtilBC.blockPosToString(min) + "] -> [" + StringUtilBC
            .blockPosToString(max) + "] @ " + StringUtilBC.blockPosToString(current) + " order: [" + order + "]"
            + (invert ? " inverting" : "") + (repeat ? " repeating" : "") + " }";
    }

    /** Moves on to the next block. Unlike {@link #next()} this returns the one AFTER that one, so you cannot use
     * {@link #hasNext()}! */
    public BlockPos advance() {
        if (current == null) {
            current = getStart();
            return getCurrent();
        }
        current = increment(current, order.first);
        if (shouldReset(order.first)) {
            if (invert) {
                order = order.invertFirst();
            }
            current = replace(current, order.first);
            current = increment(current, order.second);
            if (shouldReset(order.second)) {
                if (invert) {
                    order = order.invertSecond();
                }
                current = replace(current, order.second);
                current = increment(current, order.third);
                if (shouldReset(order.third)) {
                    if (repeat) {
                        if (invert) {
                            order = order.invertThird();
                        }
                        current = replace(current, order.third);
                        hasRepeated = true;
                    } else {
                        current = null;
                    }
                }
            }
        }
        return getCurrent();
    }

    private static BlockPos increment(BlockPos pos, EnumFacing facing) {
        int diff = facing.func_176743_c().func_179524_a();
        int value = VecUtil.getValue(pos, facing.func_176740_k()) + diff;
        return VecUtil.replaceValue(pos, facing.func_176740_k(), value);
    }

    private boolean shouldReset(EnumFacing facing) {
        int lstReturned = VecUtil.getValue(current, facing.func_176740_k());
        BlockPos goingTo = facing.func_176743_c() == AxisDirection.POSITIVE ? max : min;
        int to = VecUtil.getValue(goingTo, facing.func_176740_k());
        if (facing.func_176743_c() == AxisDirection.POSITIVE) return lstReturned > to;
        return lstReturned < to;
    }

    /** Checks to see if {@link #advance()} has ever, or could ever, return the given block position. */
    public boolean contains(BlockPos pos) {
        if (pos.func_177958_n() < min.func_177958_n() || pos.func_177958_n() > max.func_177958_n()) {
            return false;
        }
        if (pos.func_177956_o() < min.func_177956_o() || pos.func_177956_o() > max.func_177956_o()) {
            return false;
        }
        if (pos.func_177952_p() < min.func_177952_p() || pos.func_177952_p() > max.func_177952_p()) {
            return false;
        }
        return true;
    }

    /** Checks to see if {@link #advance()} will return the given block position before this repeats. */
    public boolean willVisit(BlockPos pos) {
        if (!contains(pos)) {
            return false;
        }
        if (current == null) {
            return true;
        }
        return compare(pos) < 0;
    }

    /** Checks to see if {@link #advance()} has already returned the given block position before the last repeat. */
    public boolean hasVisited(BlockPos pos) {
        if (!contains(pos)) {
            return false;
        }
        if (current == null && !hasRepeated) {
            return false;
        }
        return compare(pos) >= 0;
    }

    private int compare(BlockPos pos) {
        int cmp = compare(pos, order.third);
        if (cmp != 0) {
            return cmp;
        }
        cmp = compare(pos, order.second);
        if (cmp != 0) {
            return cmp;
        }
        return compare(pos, order.first);
    }

    private int compare(BlockPos pos, EnumFacing direction) {
        int argVal = VecUtil.getValue(pos, direction.func_176740_k());
        int currentVal = VecUtil.getValue(current, direction.func_176740_k());
        return (currentVal - argVal) * direction.func_176743_c().func_179524_a();
    }

    /** Moves this iterator so that {@link #advance()} will return the given block position next.
     * 
     * @throws IllegalArgumentException if {@link #contains(BlockPos)} doesn't return true. */
    public void moveTo(BlockPos pos) {
        if (!contains(pos)) {
            throw new IllegalArgumentException("This " + this + " doesn't contain " + pos + "!");
        }

        EnumFacing a = order.first;
        EnumFacing b = order.second;
        EnumFacing c = order.third;

        int valueA = VecUtil.getValue(pos, a.func_176740_k());
        int valueB = VecUtil.getValue(pos, b.func_176740_k());
        int valueC = VecUtil.getValue(pos, c.func_176740_k());

        int boundA = VecUtil.getValue(max, min, a);
        int boundB = VecUtil.getValue(max, min, b);
        int boundC = VecUtil.getValue(max, min, c);

        if (!invert) {
            if (valueA != boundA) {
                current = pos.func_177972_a(a.func_176734_d());
                return;
            }

            if (valueB != boundB) {
                current = pos.func_177972_a(b.func_176734_d());
                current = VecUtil.replaceValue(current, a.func_176740_k(), VecUtil.getValue(min, max, a));
                return;
            }

            if (valueC != boundC) {
                current = pos.func_177972_a(c.func_176734_d());
                current = VecUtil.replaceValue(current, a.func_176740_k(), VecUtil.getValue(min, max, a));
                current = VecUtil.replaceValue(current, b.func_176740_k(), VecUtil.getValue(min, max, b));
                return;
            }
            current = null;
            return;
        }

        if (current == null) {
            current = getStart();
        }

        int db = compare(pos, b);
        int dc = compare(pos, c);

        BlockPos size = max.func_177973_b(min);
        int sizeB = 1 + VecUtil.getValue(size, b.func_176740_k());

        if ((dc * sizeB + db) % 2 == 1) {
            order = order.invertFirst();
        }
        if (dc % 2 == 1) {
            order = order.invertSecond();
        }

        a = order.first;
        b = order.second;
        c = order.third;

        boundA = VecUtil.getValue(max, min, a);
        boundB = VecUtil.getValue(max, min, b);
        boundC = VecUtil.getValue(max, min, c);

        if (valueA != boundA) {
            current = pos.func_177972_a(order.first.func_176734_d());
        } else if (valueB != boundB) {
            current = pos.func_177972_a(order.second.func_176734_d());
            order = order.invertFirst();
        } else if (valueC != boundC) {
            current = pos.func_177972_a(order.third.func_176734_d());
            order = order.invertFirst();
            order = order.invertSecond();
        } else {
            current = null;
        }
    }

    public boolean hasFinished() {
        return current == null;
    }

    // Iterator

    @Override
    public boolean hasNext() {
        return !hasFinished();
    }

    @Override
    public BlockPos next() {
        BlockPos c = current;
        advance();
        return c;
    }
}
